# A customized DropDown allowing placing a dropdown to the side of a widget
# it is sticked to. The original .py file was released by Kivy under the
# MIT License. PR for it is on https://github.com/kivy/kivy/pull/4402
# TLDR: This particular file is under MIT License.

__all__ = ('DropDown', )

from kivy.uix.scrollview import ScrollView
from kivy.properties import ObjectProperty, NumericProperty, BooleanProperty
from kivy.core.window import Window
from kivy.lang import Builder
from kivy.clock import Clock
from kivy.config import Config

_grid_kv = '''
GridLayout:
    size_hint_y: None
    height: self.minimum_size[1]
    cols: 1
'''


class DropDownException(Exception):
    '''DropDownException class.
    '''
    pass


class DropDown(ScrollView):
    '''DropDown class. See module documentation for more information.

    :Events:
        `on_select`: data
            Fired when a selection is done. The data of the selection is passed
            in as the first argument and is what you pass in the :meth:`select`
            method as the first argument.
        `on_dismiss`:
            .. versionadded:: 1.8.0

            Fired when the DropDown is dismissed, either on selection or on
            touching outside the widget.
    '''

    auto_width = BooleanProperty(True)
    '''By default, the width of the dropdown will be the same as the width of
    the attached widget. Set to False if you want to provide your own width.
    '''

    max_height = NumericProperty(None, allownone=True)
    '''Indicate the maximum height that the dropdown can take. If None, it will
    take the maximum height available until the top or bottom of the screen
    is reached.

    :attr:`max_height` is a :class:`~kivy.properties.NumericProperty` and
    defaults to None.
    '''

    dismiss_on_select = BooleanProperty(True)
    '''By default, the dropdown will be automatically dismissed when a
    selection has been done. Set to False to prevent the dismiss.

    :attr:`dismiss_on_select` is a :class:`~kivy.properties.BooleanProperty`
    and defaults to True.
    '''

    auto_dismiss = BooleanProperty(True)
    '''By default, the dropdown will be automatically dismissed when a
    touch happens outside of it, this option allow to disable this
    feature

    :attr:`auto_dismiss` is a :class:`~kivy.properties.BooleanProperty`
    and defaults to True.

    .. versionadded:: 1.8.0
    '''

    min_state_time = NumericProperty(0)
    '''Minimum time before the :class:`~kivy.uix.DropDown` is dismissed.
    This is used to allow for the widget inside the dropdown to display
    a down state or for the :class:`~kivy.uix.DropDown` itself to
    display a animation for closing.

    :attr:`min_state_time` is a :class:`~kivy.properties.NumericProperty`
    and defaults to the `Config` value `min_state_time`.

    .. versionadded:: 1.9.2
    '''

    allow_sides = BooleanProperty(False)
    '''Property that will allow placement of a :class:`~kivy.uix.DropDown`
    on the sides of a widget set in :attr:`attach_to`.

    The default side is the right side of a widget. When there is no space
    on the right side, it'll automatically switch to the left side of the
    widget.

    :attr:`allow_sides` is a :class:`~kivy.properties.BooleanProperty`
    and defaults to False.
    '''

    attach_to = ObjectProperty(allownone=True)
    '''(internal) Property that will be set to the widget to which the
    drop down list is attached.

    The :meth:`open` method will automatically set this property whilst
    :meth:`dismiss` will set it back to None.
    '''

    container = ObjectProperty()
    '''(internal) Property that will be set to the container of the dropdown
    list. It is a :class:`~kivy.uix.gridlayout.GridLayout` by default.
    '''

    __events__ = ('on_select', 'on_dismiss')

    def __init__(self, **kwargs):
        self._win = None
        if 'min_state_time' not in kwargs:
            self.min_state_time = float(
                Config.get('graphics', 'min_state_time'))
        if 'container' not in kwargs:
            c = self.container = Builder.load_string(_grid_kv)
        else:
            c = None
        if 'allow_sides' not in kwargs:
            self.allow_sides = False
        if 'do_scroll_x' not in kwargs:
            self.do_scroll_x = False
        if 'size_hint' not in kwargs:
            if 'size_hint_x' not in kwargs:
                self.size_hint_x = None
            if 'size_hint_y' not in kwargs:
                self.size_hint_y = None
        super(DropDown, self).__init__(**kwargs)
        if c is not None:
            super(DropDown, self).add_widget(c)
            self.on_container(self, c)
        Window.bind(
            on_key_down=self.on_key_down,
            size=self._reposition)
        self.fbind('size', self._reposition)

    def on_key_down(self, instance, key, scancode, codepoint, modifiers):
        if key == 27 and self.get_parent_window():
            self.dismiss()
            return True

    def on_container(self, instance, value):
        if value is not None:
            self.container.bind(minimum_size=self._container_minimum_size)

    def open(self, widget):
        '''Open the dropdown list and attach it to a specific widget.
        Depending on the position of the widget within the window and
        the height of the dropdown, the dropdown might be above or below
        that widget.
        '''
        # ensure we are not already attached
        if self.attach_to is not None:
            self.dismiss()

        # we will attach ourself to the main window, so ensure the
        # widget we are looking for have a window
        self._win = widget.get_parent_window()
        if self._win is None:
            raise DropDownException(
                'Cannot open a dropdown list on a hidden widget')

        self.attach_to = widget
        widget.bind(pos=self._reposition, size=self._reposition)
        self._reposition()

        # attach ourself to the main window
        self._win.add_widget(self)

    def dismiss(self, *largs):
        '''Remove the dropdown widget from the window and detach it from
        the attached widget.
        '''
        Clock.schedule_once(lambda dt: self._real_dismiss(),
                            self.min_state_time)

    def _real_dismiss(self):
        if self.parent:
            self.parent.remove_widget(self)
        if self.attach_to:
            self.attach_to.unbind(pos=self._reposition, size=self._reposition)
            self.attach_to = None
        self.dispatch('on_dismiss')

    def on_dismiss(self):
        pass

    def select(self, data):
        '''Call this method to trigger the `on_select` event with the `data`
        selection. The `data` can be anything you want.
        '''
        self.dispatch('on_select', data)
        if self.dismiss_on_select:
            self.dismiss()

    def on_select(self, data):
        pass

    def _container_minimum_size(self, instance, size):
        if self.max_height:
            self.height = min(size[1], self.max_height)
            self.do_scroll_y = size[1] > self.max_height
        else:
            self.height = size[1]
            self.do_scroll_y = True

    def add_widget(self, *largs):
        if self.container:
            return self.container.add_widget(*largs)
        return super(DropDown, self).add_widget(*largs)

    def remove_widget(self, *largs):
        if self.container:
            return self.container.remove_widget(*largs)
        return super(DropDown, self).remove_widget(*largs)

    def clear_widgets(self):
        if self.container:
            return self.container.clear_widgets()
        return super(DropDown, self).clear_widgets()

    def on_touch_down(self, touch):
        if super(DropDown, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos):
            return True
        if (self.attach_to and self.attach_to.collide_point(
                *self.attach_to.to_widget(*touch.pos))):
            return True
        if self.auto_dismiss:
            self.dismiss()

    def on_touch_up(self, touch):
        if super(DropDown, self).on_touch_up(touch):
            return True
        if 'button' in touch.profile and touch.button.startswith('scroll'):
            return
        if self.collide_point(*touch.pos):
            return True
        if self.auto_dismiss:
            self.dismiss()

    def _reposition(self, *largs):
        # calculate the coordinate of the attached widget in the window
        # coordinate system
        win = self._win
        widget = self.attach_to
        if not widget or not win:
            return
        wx, wy = widget.to_window(*widget.pos)
        wright, wtop = widget.to_window(widget.right, widget.top)

        # set width and x
        if self.auto_width:
            self.width = wright - wx

        # ensure the dropdown list doesn't get out on the X axis, with a
        # preference to 0 in case the list is too wide.
        if self.allow_sides:
            if not self.auto_width:
                x = wx + widget.right
            else:
                x = wx + self.width
            if x + self.width > win.width:
                x = wx - self.width
        else:
            x = wx
            if x + self.width > win.width:
                x = win.width - self.width
        if x < 0:
            x = 0
        self.x = x

        # determine if we display the dropdown upper or lower to the widget
        h_bottom = wy - self.height
        h_top = win.height - (wtop + self.height)
        if h_bottom > 0:
            if self.allow_sides:
                self.top = wy + widget.height
            else:
                self.top = wy
            self.height = self.container.minimum_height
        elif h_top > 0:
            if self.allow_sides:
                self.y = wtop - widget.height
            else:
                self.y = wtop
            self.height = self.container.minimum_height
        else:
            # none of both top/bottom have enough place to display the
            # widget at the current size. Take the best side, and fit to
            # it.
            height = max(h_bottom, h_top)
            if height == h_bottom:
                if self.allow_sides:
                    self.top = wy + widget.height
                else:
                    self.top = wy
                self.height = wy
            else:
                if self.allow_sides:
                    self.y = wtop - widget.height
                else:
                    self.y = wtop
                self.height = win.height - wtop


if __name__ == '__main__':
    from kivy.uix.button import Button
    from kivy.base import runTouchApp

    def show_dropdown(button, *largs):
        dp = DropDown(allow_sides=True)
        dp.bind(on_select=lambda instance, x: setattr(button, 'text', x))
        for i in range(10):
            item = Button(text='hello %d' % i, size_hint_y=None, height=44)
            item.bind(on_release=lambda btn: dp.select(btn.text))
            dp.add_widget(item)
        dp.open(button)

    def touch_move(instance, touch):
        instance.center = touch.pos

    btn = Button(text='SHOW', size_hint=(None, None), pos=(300, 200))
    btn.bind(on_release=show_dropdown, on_touch_move=touch_move)

    runTouchApp(btn)



'''
.. toctree::
   mod_krysa_tasks_basic
   mod_krysa_tasks_avgs
   mod_krysa_tasks_manipulate
'''

from kivy.app import App
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.uix.popup import Popup
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.textinput import TextInput
from kivy.properties import ObjectProperty
import re
import os.path as op

Builder.load_file(op.join(op.dirname(op.abspath(__file__)), 'tasks.kv'))


class AddressLayout(BoxLayout):
    '''Simple layout that consists of single restricted input widget fetching
    only ``[a-zA-Z0-9:]`` values i.e. address.
    '''


class SmallLargeLayout(BoxLayout):
    '''A layout that consists of multiple restricted input widgets for address
    and `k` value.

    .. versionadded:: 0.1.0
    '''


class AvgsLayout(BoxLayout):
    '''A layout that consists of multiple restricted input widgets for address
    and `p` (power) value for the formula of generalized mean.

    .. versionadded:: 0.2.4
    '''


class FloatInput(TextInput):
    '''A TextInput with float filter.

    .. versionadded:: 0.3.8
    '''
    def __init__(self, **kwargs):
        super(FloatInput, self).__init__(**kwargs)
        self.input_filter = self.floatfilter

    def floatfilter(self, substring, from_undo):
        '''A function filtering everything that is not `-` symbol, floating
        point symbol(`.`) or a number.
        '''
        txt = self.text
        if '-' in txt and '.' not in txt:
            chars = re.findall(r'([0-9.])', substring)
        elif '.' in txt:
            if '-' not in txt:
                chars = re.findall(r'([\-0-9])', substring)
            else:
                chars = re.findall(r'([0-9])', substring)
        else:
            chars = re.findall(r'([\-0-9.])', substring)
        return u''.join(chars)


class FreqLayout(BoxLayout):
    '''A layout that consists of multiple checkboxes and restricted input
    widgets for address, type of values, type of output frequency and
    limits of the input values.

    .. versionadded:: 0.3.2
    '''


class SortLayout(BoxLayout):
    '''A layout that consists only of a spinner with two values:

    * Ascending
    * Descending

    The :ref:`task` with this layout is using
    :mod:`tasks.manipulate.Manipulate._manip_sort`.

    .. versionadded:: 0.3.5
    '''


class AppendLayout(BoxLayout):
    '''A layout that consists of a spinner with two values:

    * Rows
    * Columns

    and a restricted input that allows only integers.

    .. versionadded:: 0.3.6
    '''


class StandLayout(BoxLayout):
    '''(Not yet implemented)
    A layout that consists of a spinner with types of available
    standardization for values.

    * Standard score
        .. math:: X' = \\frac {X - \\mu}{\\sigma}

    * Student's t-statistic
        .. math:: X' = \\frac {X - {\\overline {X}}}{s}

    * Studentized residual
        .. math::
           \\frac {{\\hat {\\epsilon}}_{i}}{{\\hat {\\sigma}}_{i}}
           =\\frac {X_{i}-{\\hat {\\mu}}_{i}}{{\\hat {\\sigma}}_{i}}

    * Standardized moment
        .. math:: \\frac {\\mu _{k}}{\\sigma ^{k}}

    * Coefficient of variation
        .. math:: \\frac {\\sigma}{\\mu}

    * Feature scaling
        .. math:: X'=a + \\frac {(X-X_{\\min})(b-a)}{X_{\\max}-X_{\\min}}
    '''


class Task(Popup):
    '''A popup handling the basic choosing of :ref:`data` from available
    :ref:`sqlite` in the application.

    .. versionadded:: 0.1.0

    .. versionchanged:: 0.2.3
       Placed into a separated module.
    '''
    run = ObjectProperty(None)

    def __init__(self, **kw):
        '''docs
        '''
        super(Task, self).__init__(**kw)
        self.app = App.get_running_app()
        self.run = kw.get('run')
        wdg = kw.get('wdg')
        self.call = kw.get('call')
        self.from_address = self.app.root.from_address
        self.set_page = self.app.root.set_page
        if wdg:
            self.ids.container.add_widget(wdg)

    def recalc_height(self, body, content):
        '''Recalculates the height of :mod:`tasks.Task` after a layout is
        added, so that the children are clearly visible without any stretching.

        .. versionadded:: 0.3.2
        '''
        confirms = self.ids.confirms
        content.height = sum([child.height for child in content.children])
        body.height = sum([child.height for child in body.children])
        self.height = body.height + confirms.height + self.separator_height

    @staticmethod
    def get_table_pos(text, values, *args):
        '''Returns an index of selected :mod:`main.Table` from all available
        in the list.

        .. versionadded:: 0.1.0
        '''
        gen = (i for i, val in enumerate(values) if val == text)
        for i in gen:
            return i

    def try_run(self, *args):
        '''Tries to run a :ref:`task` from the input a user specified and
        closes the popup. If no such an action is possible, it'll show a popup
        with an error and leave :mod:`tasks.Task` opened.

        .. versionadded:: 0.2.0
        '''
        try:
            self.run(*args)
            if self.call:
                but = Button(size_hint_y=None, height='25dp',
                             text=self.call[0])
                but.bind(on_release=self.call[1])
                self.app.root.ids.recenttasks.add_widget(but)
            self.dismiss()
        except Exception as err:
            Logger.exception(err)
            error = self.app.errorcls(msg=repr(err))
            error.open()



# -*- coding: utf-8 -*-
# KrySA - Statistical analysis for rats
# Version: 0.3.10
# Copyright (C) 2016, KeyWeeUsr(Peter Badida) <keyweeusr@gmail.com>
# License: GNU GPL v3.0, More info in LICENSE.txt

from kivy.config import Config
import os

# local & RTD docs fix
if not any(['BUILDDIR' in os.environ, 'READTHEDOCS' in os.environ]):
    Config.set('graphics', 'window_state', 'maximized')
    import numpy
    import scipy

import re
import json
import math
import string
import sqlite3
import os.path as op
from kivy.app import App
from kivy.clock import Clock
from functools import partial
from dropdown import DropDown
from kivy.logger import Logger
from kivy.uix.popup import Popup
from kivy.uix.image import Image
from kivy.uix.label import Label
from time import gmtime, strftime
from kivy.uix.widget import Widget
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.textinput import TextInput
from kivy.uix.scrollview import ScrollView
from kivy.uix.gridlayout import GridLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.stencilview import StencilView
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.treeview import TreeView, TreeViewLabel
from kivy.uix.recyclegridlayout import RecycleGridLayout
from kivy.properties import StringProperty, ObjectProperty, \
    BooleanProperty, ListProperty

from tasks import Task
from tasks.basic import Basic
from tasks.avgs import Avgs
from tasks.manipulate import Manipulate


class ResultGrid(GridLayout):
    '''A black gridlayout, together with :mod:`main.Wrap` makes a table
    container for results that need a table.

    .. versionadded:: 0.3.2
    '''


class Wrap(Label):
    '''A white label with automatically wrapped text.

    .. versionadded:: 0.3.2
    '''
    background_color = ListProperty([0, 0, 0, 0])


class PageBox(BoxLayout):
    '''A layout that includes Page widget together with transparent separator.
    It's used for adding new results from Tasks.

    .. versionadded:: 0.2.0
    '''
    def __init__(self, **kwargs):
        super(PageBox, self).__init__(**kwargs)
        self.add_widget = self.ids.page.add_widget


class PaperLabel(Label):
    '''A label with visual properties as a paper sheet.

    .. versionadded:: 0.2.0
    '''


class ImgButton(Button):
    '''A button with an image of square shape in the middle.

    .. versionadded:: 0.2.0
    '''
    source = StringProperty('')


class ErrorPop(Popup):
    '''An error popup to let user know something is missing or typed wrong
    when console is disabled.

    .. versionadded:: 0.1.2
    '''
    message = StringProperty('')

    def __init__(self, **kw):
        self.message = kw.get('msg', '')
        super(ErrorPop, self).__init__(**kw)


class NewDataValue(BoxLayout):
    '''A layout handling the behavior of inputs and button for each new
    value in :ref:`data`.

    .. versionadded:: 0.1.4
    '''
    def __init__(self, **kw):
        self.app = App.get_running_app()
        super(NewDataValue, self).__init__(**kw)
        self.filter = kw['filter']


class NewDataColumn(BoxLayout):
    '''A layout handling the behavior of type, values(``NewDataValue``) and
    some buttons for each new column in :ref:`data`.

    .. versionadded:: 0.1.4
    '''
    def __init__(self, **kw):
        self.app = App.get_running_app()
        super(NewDataColumn, self).__init__(**kw)

    @staticmethod
    def free(items):
        '''Frees all locked cells in the column except a column type. If
        a wrong type is used, removing the whole column is necessary.
        (protection against corrupting :ref:`sqlite`)

        .. versionadded:: 0.1.4
        '''
        for item in items:
            if hasattr(item, 'disabled'):
                try:
                    if item.parent.ids['coltype'] == item:
                        continue
                except:
                    item.disabled = False
            else:
                # enable children
                for i in item:
                    try:
                        if i.parent.ids['coltype'] == item:
                            continue
                    except:
                        i.disabled = False

    def checklock(self, disable, coltype, check, *args):
        '''Disables all cells in the column, then check them against a list
        of strings that could be used to corrupt :ref:`sqlite` . If the check
        is done without an error, another check is made to protect against
        using an empty string ``''`` as a value, which if used inappropriately
        results in a crash.

        .. versionadded:: 0.1.4
        '''
        msg = 'Please remove any SQL keyword present in the column!'

        for item in disable:
            if hasattr(item, 'disabled'):
                item.disabled = True
            else:
                # disable children
                for i in item:
                    i.disabled = True

        # check for any sql keywords and correct values
        for item in check:
            try:
                for keyword in self.app.sql_blacklist:
                    if keyword in item.text.upper():
                        error = ErrorPop(msg=msg)
                        error.open()
                        return
            except AttributeError:
                for i in item.children:
                    for keyword in self.app.sql_blacklist:
                        if keyword in i.ids.value.text.upper():
                            error = ErrorPop(msg=msg)
                            error.open()
                            return
                    try:
                        if not coltype.text == 'TEXT':
                            float(i.ids.value.text)
                    except ValueError:
                        if coltype.text == 'REAL':
                            i.ids.value.text = '0.0'
                        elif coltype.text == 'INTEGER':
                            i.ids.value.text = '0'

    def paste(self, values, sep):
        '''Paste a value(s) from a user's clipboard as a column values. A user
        can choose what kind of separator was used on the values, for example::

            1 2 3 4 5     # (space)
            1\\t2\\t3\\t4\\t  # (tab)
            1\\n2\\n3\\n4\\n  # Unix-like new line character (<enter>/<return>)

        If in doubt and your values were copied from a column (e.g.
        spreadsheet), use `OS default`, which will choose between ``\\n``
        (Unix-like) or ``\\r\\n`` (Windows) new line separators.

        .. versionadded:: 0.3.4
        '''
        if sep == 'space':
            values = values.split(' ')
        elif sep == 'OS default':
            values = values.split(os.linesep)
        else:
            values = values.split(sep[1:])
        for val in values:
            v = NewDataValue(filter=self.filter)
            v.ids.value.text = val
            self.ids.vals.add_widget(v)
        self.ids.vals.parent.scroll_to(self.ids.vals.children[0])


class NewDataLayout(BoxLayout):
    '''A layout handling the behavior of ``NewDataColumn`` and some inputs for
    each new value in :ref:`data`.

    .. versionadded:: 0.1.3
    '''
    def __init__(self, **kw):
        self.app = App.get_running_app()
        super(NewDataLayout, self).__init__(**kw)


class CreateWizard(Popup):
    '''A popup handling the behavior for creating a new :ref:`data`,
    i.e a wizard.

    .. versionadded:: 0.1.3
    '''
    run = ObjectProperty(None)

    def __init__(self, **kw):
        super(CreateWizard, self).__init__(**kw)
        self.run = kw.get('run')
        wdg = kw.get('wdg')
        if wdg:
            self.ids.container.add_widget(wdg)


class Dialog(Popup):
    '''A dialog handling the behavior for creating or opening files e.g.
    :ref:`project` or :ref:`data`.

    .. versionadded:: 0.1.0
    '''
    confirm = StringProperty('')
    run = ObjectProperty(None)
    dirs = BooleanProperty(False)
    project = BooleanProperty(False)
    filter = ListProperty([])

    def __init__(self, **kw):
        super(Dialog, self).__init__(**kw)
        self.confirm = kw.get('confirm', '')
        self.run = kw.get('run')
        self.dirs = kw.get('dirs', False)
        self.project = kw.get('project', False)
        if self.project:
            self.ids.name.hint_text = 'Project.krysa'
            self.filter = [lambda folder, fname: fname.endswith('.krysa')]
        else:
            self.ids.name.hint_text = 'example.sqlite'
            self.filter = [lambda folder, fname: fname.endswith('.sqlite')]


class SideItem(BoxLayout):
    '''Supposed to be a part of settings, most likely will be removed/replaced.

    .. versionadded:: 0.1.0
    '''


class TableItem(TextInput):
    '''An item handling the behavior or each separate value in the
    :mod:`main.Table` such as updating/editing values in :ref:`data`.

    .. versionadded:: 0.1.0
    '''
    def __init__(self, **kwargs):
        super(TableItem, self).__init__(**kwargs)
        self.bind(focus=self.on_focus)

    def on_focus(self, widget, focused):
        '''Makes sure the unconfirmed value is discarded e.g. when clicked
        outside of the widget.
        '''
        if not focused:
            self.text = self.old_text

    def update_value(self, txt, *args):
        '''On ``<enter>`` (``return``) key updates the values
        :mod:`main.TableItem.text` and :mod:`main.TableItem.old_text` in
        :mod:`main.Table`.

        .. versionadded:: 0.1.0
        '''
        data = []
        cols = self.cols - 1

        for i in self.origin.data:
            if 'cell' in i:
                data.append(i)
        chunks = [data[x:x + cols] for x in xrange(0, len(data), cols)]

        orig_type = type(chunks[self.r][self.c - 1])
        place = self.cols * (self.r + 1) - (self.cols - self.c)
        self.origin.data[place]['text'] = txt
        self.origin.data[place]['old_text'] = txt
        self.origin.refresh_from_data()


class Table(ScrollView):
    '''A view handling the behavior of the inputs from :ref:`sqlite`. Separates
    the values from :ref:`sqlite` according to its :ref:`data`'s column types
    into three Python categories - `int`, `float` or `unicode` and assigns
    an alphabetic order for each column together with row number to each value.

    .. versionadded:: 0.1.0
    '''
    # use with ....add_widget(Table(max_cols=3, max_rows=3))
    # Grid -> Scroll, grid as container - better for sizing and placing

    # set "address" for table pos in grid to item
    number_size = (30, 30)
    default_size = (100, 30)

    def __init__(self, **kw):
        app = App.get_running_app()
        self.rv = RecycleView(bar_width='10dp',
                              scroll_type=['bars', 'content'])
        container = RecycleGridLayout(size_hint=(None, None))
        container.viewclass = TableItem
        container.bind(minimum_size=container.setter('size'))

        self.max_rows = kw.get('max_rows', 1)
        self.rows = self.max_rows + 1
        self.max_cols = kw.get('max_cols', 1)
        self.cols = self.max_cols + 1
        self.values = kw.get('values', [])
        self.labels = kw.get('labels')
        self.types = kw.get('types')

        container.rows = self.rows
        container.cols = self.cols

        ltr = [' (' + letter + ')' for letter in self.get_letters()]
        self.rv.add_widget(container)

        if 'item_size' not in kw:
            item_size = self.default_size
        super(Table, self).__init__(**kw)
        for r in range(self.rows):
            for c in range(self.cols):
                if r == 0:
                    if c == 0:
                        self.rv.data.append({'text': u'', 'disabled': True,
                                             'size': self.number_size,
                                             'origin': self.rv})
                    else:
                        self.rv.data.append(
                            {
                                'text': self.labels[c - 1] + ltr[c - 1],
                                '_text': self.labels[c - 1],
                                'disabled': True,
                                'cell': 'label' + str(c - 1),
                                'type': type(u''),
                                'size': self.default_size,
                                'origin': self.rv
                            }
                        )
                else:
                    if c == 0:
                        self.rv.data.append({'text': unicode(r),
                                             'disabled': True,
                                             'size': self.number_size,
                                             'origin': self.rv})
                    else:
                        filter = app.root.simple_chars
                        try:
                            val = self.values.pop(0)
                            text_type = type(val)
                            filter_val = repr(text_type)[7:-2]
                            if filter_val in ['int', 'float']:
                                filter = filter_val
                        except IndexError:
                            print 'values < space'
                            val = '.'

                        if 'e+' in str(val) or 'e-' in str(val):
                            val = '{0:.10f}'.format(val)
                        if self.types:
                            if self.types[c - 1] == 'INTEGER':
                                filter = 'int'
                                text_type = type(1)
                            elif self.types[c - 1] == 'REAL':
                                filter = 'float'
                                text_type = type(1.1)
                        self.rv.data.append(
                            {
                                'text': unicode(val),
                                'old_text': unicode(val),
                                'disabled': False,
                                'cell': self.labels[c - 1] + str(r),
                                'r': r,
                                'rows': self.rows,
                                'c': c,
                                'cols': self.cols,
                                'type': text_type,
                                'size': self.default_size,
                                'input_filter': filter,
                                'origin': self.rv
                            }
                        )
        if self.values:
            raise Exception('Not enough space for all values! \
                            Increase rows/cols!')
        self.add_widget(self.rv)

    def get_letters(self):
        '''Gets a list of letters the same length as :ref:`data`'s columns.

        .. versionadded:: 0.1.0
        '''
        letters = [chr(letter + 65) for letter in range(26)]
        result = []
        label = []
        cols = range(self.cols + 1)
        for i in cols:
            if i != 0:
                while i:
                    i, rem = divmod(i - 1, 26)
                    label[:0] = letters[rem]
                result.append(''.join(label))
                label = []
        return result

    def lock(self, disabled=True):
        '''docs

        .. versionadded:: 0.1.0
        '''
        for i in self.rv.data:
            if 'cell' in i and 'label' not in i['cell']:
                i['disabled'] = disabled
        self.rv.refresh_from_data()

    def clean(self, *args):
        '''Removes all data from :mod:`main.Table`

        .. versionadded:: 0.1.0
        '''
        self.rv.data = []


class ProcessFlow(BoxLayout, StencilView):
    '''A canvas on which will be displayed actions for each :ref:`data` related
    to them, such as used tasks connected with result of the tasks.

    .. versionadded:: 0.1.0

    (Not implemented yet)
    '''
    def __init__(self, **kw):
        super(ProcessFlow, self).__init__(**kw)
        app = App.get_running_app()
        self.texture = Image(source=app.path + '/data/grid.png').texture
        self.texture.wrap = 'repeat'


class SizedButton(Button):
    '''A button with width automatically customized according to text length of
    its siblings, which makes every sibling the same size as the one with the
    longest text string.

    .. versionadded:: 0.1.0
    '''
    def correct_width(self, *args):
        self.width = self.texture_size[0] + 8
        self.parent.parent.width = max([c.width for c in self.parent.children])
        for child in self.parent.children:
            child.width = self.parent.parent.width


class MenuDrop(DropDown):
    '''A list of :mod:`main.SizedButton` s displayed as a menu, where each
    button may create another menu depending on the function bound to it. The
    main menu is handled through a single instance of :mod:`main.MenuDrop`
    which is instantiated before :mod:`main.Krysa.build` function.

    Each click/tap on the menu button then assigns a value to it from
    ``App.menu`` dictionary according to its name in `kv` file.

    .. versionadded:: 0.1.0
    '''
    def __init__(self, **kw):
        app = App.get_running_app()
        app.drop = self
        super(MenuDrop, self).__init__(**kw)

    def click(self, instance, values):
        for value in values:
            btn = SizedButton(text=value[0])
            btn.bind(on_release=value[1])
            self.add_widget(btn)
        self.open(instance)

    def on_dismiss(self, *args):
        self.clear_widgets()


class Body(FloatLayout):
    '''The main layout for the application. It handles menu values, their
    appropriate functions, filtering of user's input and functions for
    accessing :ref:`sqlite` in :mod:`main.Table`.

    .. versionadded:: 0.1.0
    '''
    def __init__(self, **kw):
        self.app = App.get_running_app()
        self.tables = []
        self.app.menu = {'file': (['New...', self.new],
                                  ['Open Project', self.open_project],
                                  ['Close Project', self.close_project],
                                  ['Save Project', self.save_project],
                                  ['Import Data', self.import_data],
                                  ['Export Data', self.export_data],
                                  ['_Recent Projects', self.test],
                                  ['Exit', self.app.stop],),
                         'edit': (['_Undo', self.test],
                                  ['_Redo', self.test],
                                  ['_Cut', self.test],
                                  ['_Copy', self.test],
                                  ['_Paste', self.test],
                                  ['_Delete', self.test],
                                  ['_Find...', self.test],
                                  ['_Replace...', self.test],
                                  ['_Protect Data', self.test],
                                  ['_Go To...', self.test],),
                         'tasks': (['Basic', self.basic],
                                   ['Averages', self.avgs],
                                   ['Manipulate', self.manipulate],),
                         'help': (['_KrySA Help', self.test],
                                  ['_Getting Started Tutorial', self.test],
                                  ['About KrySA', self.about],),
                         }
        super(Body, self).__init__(**kw)

    def new(self, button, *args):
        '''Opens a submenu for ``New`` menu.

        .. versionadded:: 0.1.0
        '''
        d = DropDown(allow_sides=True, auto_width=False)
        buttons = []

        # Project saving works, but implementing ProcessFlow is still ToDo
        buttons.append(SizedButton(text='Project'))
        buttons[0].bind(on_release=self._new_project)
        buttons.append(SizedButton(text='Data'))
        buttons[1].bind(on_release=self._new_data)
        for b in buttons:
            d.add_widget(b)
        d.open(button)

    def _new_project(self, *args):
        '''Closes already opened :ref:`project` if available and opens a dialog
        for creating a new one.

        .. versionadded:: 0.1.2
        '''
        self.close_project()
        self.savedlg = Dialog(title='New Project',
                              confirm='Save',
                              run=self._save_project,
                              dirs=True,
                              project=True)
        self.savedlg.open()

    def _new_data(self, *args):
        '''Opens a wizard for creating a new :ref:`data` if a :ref:`project` is
        available or shows a warning if it doesn't exist.

        .. versionadded:: 0.1.3
        '''
        if not self.app.project_exists:
            error = ErrorPop(msg='No project exists!')
            error.open()
            return
        widget = NewDataLayout()
        self.wiz_newdata = CreateWizard(title='New Data', wdg=widget,
                                        run=partial(self._save_data, widget))
        self.wiz_newdata.open()

    def _save_data(self, wizard, *args):
        '''Gets data from the wizard, puts them into :mod:`main.Table` and exports them
        into :ref:`sqlite`.

        .. versionadded:: 0.1.4
        '''
        labels = []
        types = []
        values = []
        table_name = wizard.ids.table_name.text

        if not len(table_name):
            error = ErrorPop(msg='Please name your data!')
            error.open()
            return
        if not len(wizard.ids.columns.children):
            error = ErrorPop(msg='There must be at least one column!')
            error.open()
            return

        for child in reversed(wizard.ids.columns.children):
            child.ids.checklock.dispatch('on_release')
        for child in reversed(wizard.ids.columns.children):
            if child.ids.colname.text not in labels:
                lbl = re.findall(r'([a-zA-Z0-9])', child.ids.colname.text)
                if not len(lbl):
                    error = ErrorPop(msg='Use only a-z A-Z 0-9 characters!')
                    error.open()
                    return
                labels.append(''.join(lbl))
            else:
                error = ErrorPop(msg='Each column must have a unique name!')
                error.open()
                return

            types.append(child.ids.coltype.text)
            column_values = []

            for value_wdg in reversed(child.ids.vals.children):
                column_values.append(value_wdg.ids.value.text)
            values.append(column_values)

        max_cols = len(values)
        max_rows = len(max(values, key=len))

        # create table in KrySA, then export
        tabletab = TabbedPanelItem(text=table_name)
        self.ids.tabpanel.add_widget(tabletab)
        _values = values[:]
        values = []

        for i in range(max_rows):
            for j in range(max_cols):
                try:
                    values.append(_values[j][i])
                except IndexError:
                    if types[j] == 'INTEGER':
                        values.append(0)
                    elif types[j] == 'REAL':
                        values.append(0.0)
                    else:
                        values.append(u'')

        self.tables.append((table_name, Table(max_cols=max_cols,
                                              max_rows=max_rows, pos=self.pos,
                                              size=self.size, values=values,
                                              labels=labels, types=types)))
        tabletab.content = self.tables[-1][1]

        # export to data.sqlite in <project>/data directory
        data = op.join(self.app.project_dir, 'data')

        self._export_data([data], 'data.sqlite')
        self.wiz_newdata.dismiss()

    def open_project(self, *args):
        self.close_project()
        self.opendlg = Dialog(title='Open Project',
                              confirm='Open',
                              run=self._open_project,
                              project=True)
        self.opendlg.open()

    def _open_project(self, selection, *args):
        '''Opens a :ref:`project` from path selected in ``Dialog`` and imports
        :ref:`sqlite`.

        .. versionadded:: 0.1.7
        '''
        if not selection:
            return
        else:
            selection, fname = op.split(selection[0])

        data = op.join(selection, 'data')
        results = op.join(selection, 'results')

        self.app.project_exists = True
        self.app.project_dir = selection
        self.app.project_name = fname.split('.')[0]

        # (dummy for now)
        # dump widgets' properties from process flow to dict, then to json
        with open(op.join(selection, fname), 'rb') as f:
            project = json.loads(f.read())
        print project

        # import from project automatically
        self._import_data([op.join(data, 'data.sqlite')])

        # import results
        for file in os.listdir(results):
            self.set_page('', op.join(results, file), result_type='import')

    def close_project(self, *args):
        '''Clears all important variables, removes all :ref:`data` available in
        :mod:`main.Table` and switches to :mod:`main.ProcessFlow`.

        .. versionadded:: 0.1.0
        '''
        # call this before a new project
        self.app.project_exists = False
        self.app.project_dir = ''
        self.app.project_name = ''
        tp = self.ids.tabpanel
        while len(tp.tab_list) > 1:
            self.tables.pop()
            tp.remove_widget(tp.tab_list[0])
        self.ids.flow.clear_widgets()
        tp.switch_to(tp.tab_list[0])

    def save_project(self, *args):
        # same as _new_project + export data & results

        # make some output if unsaved changes available(e.g. * before name)
        if self.app.project_exists:
            self._save_project()
            return
        self.savedlg = Dialog(title='New Project',
                              confirm='Save',
                              run=self._save_project,
                              dirs=True,
                              project=True)
        self.savedlg.open()

    def _save_project(self, selection=None, fname=None, *args):
        '''Saves a :ref:`project` to path selected in ``Dialog`` and exports
        :ref:`sqlite`.

        .. versionadded:: 0.1.2
        '''
        if not selection:
            if not self.app.project_exists:
                return
            selection = op.dirname(self.app.project_dir)
            fname = self.app.project_name
        else:
            selection = selection[0]
        if '.' not in fname:
            fname = fname + '.krysa'

        if selection != op.dirname(self.app.project_dir):
            try:
                os.mkdir(op.join(selection, fname.split('.')[0]))
            except OSError:
                error = ErrorPop(msg='Project folder already exists!')
                error.open()
                return

        selection = op.join(selection, fname.split('.')[0])
        data = op.join(selection, 'data')
        results = op.join(selection, 'results')
        if selection != self.app.project_dir:
            os.mkdir(data)
            os.mkdir(results)
        if op.exists(op.join(selection, fname)):
            os.remove(op.join(selection, fname))

        self.app.project_exists = True
        self.app.project_dir = selection
        self.app.project_name = fname.split('.')[0]

        # (dummy for now)
        # dump widgets' properties from process flow to dict, then to json
        project = {u'test': u'blah'}
        with open(op.join(selection, fname), 'wb') as f:
            f.write(json.dumps(project, indent=4))

        # let user set table columns, add to tab, then:
        self._export_data([data], 'data.sqlite')

        # exporting results to pdf -> 0.3.x
        self._export_results(results)

    def import_data(self, *args):
        self.opendlg = Dialog(title='Import Data',
                              confirm='Import',
                              run=self._import_data)
        self.opendlg.open()

    def _import_data(self, selection, *args):
        '''Imports :ref:`sqlite` from path selected in ``Dialog`` and puts it
        to :mod:`main.Table`.

        .. versionadded:: 0.1.0
        '''
        # limit table name and column name to [a-zA-Z]

        # CREATE TABLE test(
        #                   Column INTEGER NOT NULL CHECK(
        #                               typeof(Column) = 'integer'))
        if not selection:
            return
        else:
            selection = selection[0]
            if '.sqlite' not in selection:
                return

        conn = sqlite3.connect(op.join(selection))
        c = conn.cursor()

        # get tables first!
        c.execute("SELECT name FROM sqlite_master WHERE type='table'")
        tables = [tab[0] for tab in c.fetchall()]

        for table in tables:
            c.execute("pragma table_info({0})".format(table))
            table_info = c.fetchall()
            labels = [lbl[1] for lbl in table_info]

            # allow only: INTEGER, REAL, TEXT
            try:
                types = [type[2][0] for type in table_info]
            except IndexError:
                error = ErrorPop(msg='Bad file: No defined types in columns!')
                error.open()
                return

            tabletab = TabbedPanelItem(text=table)
            self.ids.tabpanel.add_widget(tabletab)
            c.execute('select * from {0}'.format(table))
            values = [item for item in c.fetchone()]
            max_cols = len(values)
            values += [item for sublist in c.fetchall() for item in sublist]
            max_rows = int(math.ceil(len(values) / float(max_cols)))
            self.tables.append((table, Table(max_cols=max_cols,
                                             max_rows=max_rows, pos=self.pos,
                                             size=self.size, values=values,
                                             labels=labels)))
            tabletab.content = self.tables[-1][1]
        self.opendlg.dismiss()
        conn.close()
        # ToDo: implement automatic saving to Project's data.sqlite
        # ToDo: if the same table name exist in tables, add ' (2)' or something

    def export_data(self, *args):
        self.savedlg = Dialog(title='Export Data',
                              confirm='Export',
                              run=self._export_data,
                              dirs=True)
        self.savedlg.open()

    @staticmethod
    def _extract_rows(data):
        '''Extract values from :mod:`main.Table`'s dictionary into a flat list.

        Example:

        ===== ===== =====
        Data1 Data2 Data3
          1    2.0    3
        ===== ===== =====

        [u'Data1', u'Data2', u'Data3', u'1', 2.0, 3, ...]

        .. versionadded:: 0.1.0
        '''
        rows = []
        for item in data:
            try:
                item['cell']
                if issubclass(item['type'], float):
                    rows.append(float(item['text']))
                elif issubclass(item['type'], int):
                    rows.append(int(item['text']))
                else:
                    rows.append(item['text'])
            except KeyError:
                pass

            # protect against '' values -> int(''), float('')
            except ValueError:
                # duplicated zeros to prevent sqlite3.IntegrityError
                # - raised when there is type checking for inserting
                # values to the table e.g. type(0) != type(0.0)
                if issubclass(item['type'], float):
                    rows.append(0.0)
                elif issubclass(item['type'], int):
                    rows.append(0)

        data = []
        return rows

    def _export_data(self, selection, fname, *args):
        '''Exports all available :ref:`data` (visible as tabs) as :ref:`sqlite`
        into path selected in ``Dialog``.

        .. versionadded:: 0.1.1
        '''
        col_types = {"<type 'int'>": 'INTEGER', "<type 'float'>": 'REAL'}
        if not selection:
            return
        else:
            selection = selection[0]
        if op.exists(op.join(selection, fname)):
            os.remove(op.join(selection, fname))

        conn = sqlite3.connect(op.join(selection, fname))
        c = conn.cursor()

        for table in self.tables:
            col_string = ''
            max_cols = table[1].max_cols
            columns = table[1].rv.data[1:max_cols + 1]  # types need whole dict
            types = table[1].rv.data[max_cols + 2:2 * (max_cols + 1)]

            for i, col in enumerate(columns):
                try:
                    type = col_types[repr(types[i]['type'])]
                except KeyError:
                    type = 'TEXT'
                # force column to check inserted values
                col_string += (
                    col['_text'] + " " + type +
                    " NOT NULL CHECK(typeof(" + col['_text'] +
                    ") = '" + type.lower() + "'),"
                )

            if col_string.endswith(','):
                col_string = col_string[:-1]
            c.execute(
                "CREATE TABLE " + table[0] + "(" + col_string + ")"
            )
            rows = self._extract_rows(table[1].rv.data)[max_cols:]

            cnks = []
            for x in xrange(0, len(rows), max_cols):
                cnks.append(rows[x:x + max_cols])
            for chunk in cnks:
                _chunk = []
                for cnk in chunk:
                    if not isinstance(cnk, (str, unicode)):
                        _chunk.append(unicode(cnk))
                    else:
                        _chunk.append('\'' + cnk + '\'')
                values = ','.join(_chunk)
                c.execute(
                    "INSERT INTO " + table[0] +
                    " VALUES(" + values + ")"
                )
            conn.commit()
        conn.close()

        try:
            self.savedlg.dismiss()
        except AttributeError:
            # either closed, or wasn't even opened
            pass

    def _export_results(self, selection, *args):
        if not selection:
            return
        for file in os.listdir(selection):
            os.remove(op.join(selection, file))
        result_wdg = self.ids.results
        for i, result in enumerate(reversed(self.ids.results.children)):
            try:
                where = op.join(selection, str(i).zfill(3) + '.png')
                result.children[2].children[0].export_to_png(where)
            except IndexError:
                print 'No results available.'

    # menu showing functions
    @staticmethod
    def basic(button, *args):
        drop = DropDown(allow_sides=True, auto_width=False)
        for t in Basic.names:
            but = SizedButton(text=t[0])
            but.bind(on_release=t[1])
            drop.add_widget(but)
        drop.open(button)

    @staticmethod
    def avgs(button, *args):
        drop = DropDown(allow_sides=True, auto_width=False)
        for t in Avgs.names:
            but = SizedButton(text=t[0])
            but.bind(on_release=t[1])
            drop.add_widget(but)
        drop.open(button)

    @staticmethod
    def manipulate(button, *args):
        drop = DropDown(allow_sides=True, auto_width=False)
        for t in Manipulate.names:
            but = SizedButton(text=t[0])
            but.bind(on_release=t[1])
            drop.add_widget(but)
        drop.open(button)

    @staticmethod
    def about(*args):
        '''Displays `about` page of the app and includes other credits.

        .. versionadded:: 0.1.0
        '''
        aboutdlg = Popup(title='About')
        text = (
            'Copyright (C) 2016, KeyWeeUsr(Peter Badida)\n'
            'License: GNU GPL v3.0\n'
            'Find me @ https://github.com/KeyWeeUsr\n\n'
            'Used Software:\n\n'
            'Python\n'
            'Copyright (C) 2001-2016 Python Software Foundation\n'
            'All rights reserved.\n\n'
            'Kivy\n'
            'Copyright (C) 2010-2016 Kivy Team and other contributors\n\n'
            'NumPy\n'
            'Copyright (C) 2005-2016, NumPy Developers.\n'
            'All rights reserved.\n\n'
            'SciPy\n'
            'Copyright (C) 2003-2013 SciPy Developers.\n'
            'All rights reserved.\n\n'
            'MatPlotLib\n'
            'Copyright (C) 2002-2016 John D. Hunter\n'
            'All Rights Reserved\n'
        )

        aboutdlg.content = Label(text=text)
        aboutdlg.open()

    @staticmethod
    def test(*args):
        print 'ping: ', args

    # non-menu related functions
    @staticmethod
    def get_column(address):
        col = 0
        for c in address:
            if c in string.ascii_letters:
                col = col * 26 + (ord(c.upper()) - ord('A')) + 1
        return col

    def from_address(self, table, address, extended=False, *args):
        '''Gets value(s) from :mod:`main.Table` according to the address such as
        ``A1`` or ``A1:B2``. Values are fetched in the way that the final list
        contains even empty (``u''``) values. It is not expected of user to use
        :ref:`task` for strings and most of them won't even run. To get
        non-empty values for a :ref:`task` use for example Python's
        ``filter()``::

            values = filter(lambda x: len(str(x)), values)

        This `filter`, however, will remain values such as ``None`` untouched.

        .. versionadded:: 0.1.0

        .. versionchanged:: 0.3.5
           Added extended options and a possibility to get ``:all`` values from
           data.
        '''
        values = []
        col_row = []  # [column, row] such as [x, y] |_
        if ':' not in address:
            match = re.findall(r'([a-zA-Z]+)([0-9]+)', address)
            col_row.append([self.get_column(match[0][0]), match[0][1]])
        else:
            if 'all' in address:
                start_col = 1
                start_row = 1
                end_col = self.tables[table][1].cols
                end_row = self.tables[table][1].rows
            else:
                addresses = address.split(':')
                match = re.findall(r'([a-zA-Z]+)([0-9]+)', ' '.join(addresses))
                start_col = self.get_column(match[0][0])
                start_row = int(match[0][1])
                end_col = self.get_column(match[1][0])
                end_row = int(match[1][1])
            for col in range(start_col - 1, end_col):
                for row in range(start_row - 1, end_row):
                    col_row.append([col + 1, row + 1])

        col_range = 0
        row_range = 0
        val_place = [None, None]  # [column, row]
        table_cols = self.tables[table][1].cols
        table_rows = self.tables[table][1].rows
        labels = self.tables[table][1].labels

        while col_row:
            c, r = col_row.pop(0)
            c = int(c)
            r = int(r)

            if 0 < r < table_rows and 0 < c < table_cols:
                # increment ranges
                if extended:
                    if val_place[0] != c:
                        val_place[0] = c
                        col_range += 1
                    if val_place[1] != r:
                        val_place[1] = r
                        row_range += 1

                # get values
                key = table_cols * (r + 1) - (table_cols - c)
                item = self.tables[table][1].rv.data[key]
                if item['r'] == r and item['c'] == c:
                    if issubclass(item['type'], float):
                        values.append(float(item['text']))
                    elif issubclass(item['type'], int):
                        values.append(int(item['text']))
                    else:
                        values.append(item['text'])
        if extended:
            if row_range < col_range:
                # return 1 manually if only one row
                # otherwise row_range == 0
                row_range = 1
            else:
                row_range = row_range / int(col_range)
            return values, col_range, row_range, labels
        else:
            return values

    def set_page(self, task, result, result_type='text', footer='time'):
        '''Creates a :mod:`main.PageBox` for a result. The header consists of the
        :ref:`task`'s name, the footer is by default the time when the result
        was created and the content depends on `result_type` which can be -
        text, image(path to image) or widget. If `result_type == 'widget'`,
        result has to be an instance of a widget (obviously containing the
        output), e.g.::

            b = Button(text='my output')
            set_page('MyTask', b, result_type='widget')

        .. note:: When exporting pages, everything is converted into images
           (pngs), therefore making fancy behaving widgets is irrelevant.

        .. versionadded:: 0.2.0

        .. versionchanged:: 0.3.2
           Added tables as a result type.
        '''
        page = PageBox()
        head = PaperLabel(text=task, size_hint_y=None, height='30dp')

        if result_type == 'text':
            content = PaperLabel(text=result)
        elif result_type in ['image', 'import']:
            content = Image(source=result)
        elif 'table' in result_type:
            # result == list of values
            box = BoxLayout(orientation='vertical')
            grid = ResultGrid(cols=int(result_type.strip('table')),
                              padding=1, spacing=1, size_hint=[0.95, None],
                              pos_hint={'center_x': 0.5})
            grid.bind(minimum_height=grid.setter('height'))

            for value in result:
                grid.add_widget(Wrap(text=' ' + str(value), color=[0, 0, 0, 1],
                                     background_color=[1, 1, 1, 1]))

            box.add_widget(Widget())
            box.add_widget(grid)
            box.add_widget(Widget())
            content = box

        elif result_type == 'widget':
            content = result

        # turn off with footer=None
        if result_type == 'import':
            pass
        elif not footer:
            foot = PaperLabel(size_hint_y=None, height='30dp')
        else:
            if footer == 'time':
                t = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
                foot = PaperLabel(text=t,
                                  size_hint_y=None,
                                  height='30dp')
            else:
                foot = PaperLabel(text=footer, size_hint_y=None,
                                  height='30dp')

        if not result_type == 'import':
            page.add_widget(head)
            page.add_widget(content)
            page.add_widget(foot)
        else:
            page.add_widget(content)
        self.ids.results.add_widget(page, 1)

        panel = self.ids.resultspanel
        if panel.width == panel.min_size:
            panel.width = panel.max_size

    @staticmethod
    def simple_chars(substring, from_undo):
        chars = re.findall(r'([a-zA-Z0-9.])', substring)
        return u''.join(chars)

    @staticmethod
    def address_chars(substring, from_undo):
        chars = re.findall(r'([a-zA-Z0-9:])', substring)
        return u''.join(chars)

    def update_tree(self, dt):
        tree = self.ids.tree
        lab = TreeViewLabel
        for node in tree.iterate_all_nodes():
            tree.remove_node(node)
        root = tree.add_node(lab(text=self.app.project_name, is_open=True))
        for path, folders, files in os.walk(self.app.project_dir):
            if op.basename(self.app.project_dir) != op.basename(path):
                dirname = op.basename(path).title()
                parent = tree.add_node(lab(text=dirname, is_open=True), root)
                for file in files:
                    filename = op.basename(file)
                    tree.add_node(lab(text=filename, is_open=True), parent)


class KrySA(App):
    '''The main class of the application through which is handled the
    communication of other classes with getting an instance of the app via
    ``App.get_running_app()``.

    Other than that, it holds important variables of :ref:`project`, sql
    blacklist for :ref:`sqlite` creating and updating or the application
    properties themselves.
    '''
    path = op.dirname(op.abspath(__file__))
    icon = path + '/data/icon.png'
    project_exists = BooleanProperty(False)
    project_name = ''
    project_dir = ''
    title = 'KrySA'
    errorcls = ErrorPop
    tablecls = Table
    sql_blacklist = ['DROP', 'EXEC', 'DECLARE', 'UPDATE', 'CREATE', 'DELETE',
                     'INSERT', 'JOIN', '=', '"', "'", ';']

    def on_project_exists(self, instance, exists):
        '''Checks change of :mod:`main.KrySA.project_exists` and if
        :ref:`project` exists, schedules updating of its tree to 5 second
        interval.

        .. versionadded:: 0.3.0
        '''
        if exists:
            self.treeclock = Clock.schedule_interval(self.root.update_tree, 5)
        else:
            Clock.unschedule(self.treeclock)

    def build(self):
        '''Default Kivy function for getting the root widget of application.
        '''
        # read MenuDrop if an idea of removing it comes up
        MenuDrop()
        return Body()

if __name__ == '__main__':
    KrySA().run()



import unittest

import os
import sys
import time
import sqlite3
import os.path as op
from shutil import rmtree
from functools import partial
from kivy.clock import Clock

main_path = op.dirname(op.dirname(op.abspath(__file__)))
sys.path.append(main_path)
from main import KrySA


class Test(unittest.TestCase):
    def pause(*args):
        time.sleep(0.000001)

    def run_test(self, app, *args):
        Clock.schedule_interval(self.pause, 0.000001)

        # test saving empty Project
        app.root._save_project()
        project_folder = op.join(self.path, 'test_folder', 'Test')
        self.assertFalse(op.exists(project_folder))

        # open New -> Project popup, set inputs
        app.root._new_project()
        app.root.savedlg.view.selection = [self.folder, ]
        app.root.savedlg.ids.name.text = 'Test.krysa'
        app.root.savedlg.run([self.folder, ], 'Test.krysa')
        data = op.join(project_folder, 'data')
        results = op.join(project_folder, 'results')

        # open New -> Data popup, set inputs
        app.root._new_data()
        new_data = app.root.wiz_newdata.ids.container.children[0]
        new_data.ids.table_name.text = 'NewData'
        cols = new_data.ids.columns.children

        # set columns for new data
        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'INTEGER'
        vals = cols[0].ids.vals.children
        for i in range(5):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 4:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = '1'

        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'TEXT'
        vals = cols[0].ids.vals.children
        for i in range(3):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 2:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = 'end'

        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'REAL'
        vals = cols[0].ids.vals.children
        for i in range(4):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 3:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = '1.1'
        new_data = app.root.wiz_newdata.run()

        # set testing data
        newdata_values = [[0, u'', 0.0],
                          [0, u'', 0.0],
                          [0, u'end', 0.0],
                          [0, u'', 1.1],
                          [1, u'', 0.0]]

        # get data from krysa table
        end_pos = None
        ndv = []
        for d in app.root.tables[0][1].rv.data:
            if 'c' in d.keys():
                if d['type'] == int:
                    ndv.append(int(d['text']))
                elif d['type'] == float:
                    ndv.append(float(d['text']))
                else:
                    ndv.append(d['text'])
            if not end_pos and d['text'] == u'end':
                end_pos = app.root.tables[0][1].rv.data.index(d)
        ndv = [ndv[x:x + 3] for x in xrange(0, len(ndv), 3)]
        self.assertEqual(len(app.root.tables), 1)

        # test data
        self.assertTrue(op.exists(op.join(data, 'data.sqlite')))
        conn = sqlite3.connect(op.join(data, 'data.sqlite'))
        c = conn.cursor()
        c.execute('SELECT * FROM NewData')
        values = [item for sublist in c.fetchall() for item in sublist]
        values = [values[x:x + 3] for x in xrange(0, len(values), 3)]
        try:
            c.execute('SELECT * FROM NewData2')
        except sqlite3.OperationalError:
            print 'No NewData2 check... OK\n'
        conn.close()
        print 'NewData:'
        for i, v in enumerate(values):
            # values == sql values
            self.assertEqual(v, newdata_values[i])
            # values == KrySA values
            self.assertEqual(ndv[i], newdata_values[i])
            print v

        # change value in table
        app.root.tables[0][1].rv.data[end_pos]['text'] = u'new_end'
        ndv[2][1] = u'new_end'
        newdata_values[2][1] = u'new_end'
        app.root._save_project()
        print app.project_dir, app.project_exists
        # test new data
        conn = sqlite3.connect(op.join(data, 'data.sqlite'))
        c = conn.cursor()
        c.execute('SELECT * FROM NewData')
        values = [item for sublist in c.fetchall() for item in sublist]
        values = [values[x:x + 3] for x in xrange(0, len(values), 3)]
        conn.close()
        print 'NewData:'
        for i, v in enumerate(values):
            # values == sql values
            self.assertEqual(v, newdata_values[i])
            # values == KrySA values
            self.assertEqual(ndv[i], newdata_values[i])
            print v

        app.stop()

    def test_file_saveproject(self):
        self.path = op.dirname(op.abspath(__file__))
        if not op.exists(op.join(self.path, 'test_folder')):
            os.mkdir(op.join(self.path, 'test_folder'))
        else:
            rmtree(op.join(self.path, 'test_folder'))
            os.mkdir(op.join(self.path, 'test_folder'))
        self.folder = op.join(self.path, 'test_folder')

        app = KrySA()
        p = partial(self.run_test, app)
        Clock.schedule_once(p, .000001)
        app.run()
        rmtree(self.folder)

if __name__ == '__main__':
    unittest.main()



'''
.. |b| replace:: bins
.. _b: http://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram.html
'''

from . import Task, AddressLayout, SmallLargeLayout, FreqLayout
from functools import partial
from numpy import histogram
import math


class Basic(object):
    '''All :ref:`Task` s categorized as `basic` under one roof.

    .. versionadded:: 0.1.0
    '''

    def basic_count(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.AddressLayout` that
        gets from user :ref:`Data` address.

        .. versionadded:: 0.1.0
        '''
        widget = AddressLayout()
        task = Task(title='Count', wdg=widget,
                    call=['Count', Basic.basic_count])
        task.run = partial(Basic._basic_count,
                           task,
                           task.ids.container.children[0].ids.name)
        task.open()

    @staticmethod
    def _basic_count(task, address, *args):
        '''Gets the values from address and returns the count.

        .. versionadded:: 0.1.0
        '''
        values = task.from_address(task.tablenum, address.text)
        task.set_page('Count', str(len(values)), 'text')

    def basic_countifs(self, *args):
        '''Not yet implemented.
        '''

    def basic_min(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.AddressLayout` that
        gets from user :ref:`Data` address.

        .. versionadded:: 0.1.0
        '''
        widget = AddressLayout()
        task = Task(title='Minimum', wdg=widget,
                    call=['Minimum', Basic.basic_min])
        task.run = partial(Basic._basic_min,
                           task,
                           task.ids.container.children[0].ids.name)
        task.open()

    @staticmethod
    def _basic_min(task, address, *args):
        '''Gets the values from address and returns a minimum.

        .. versionadded:: 0.1.0
        '''
        values = task.from_address(task.tablenum, address.text)
        task.set_page('Minimum', str(min(values)), 'text')

    def basic_max(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.AddressLayout` that
        gets from user :ref:`Data` address.

        .. versionadded:: 0.1.0
        '''
        widget = AddressLayout()
        task = Task(title='Maximum', wdg=widget,
                    call=['Maximum', Basic.basic_max])
        task.run = partial(Basic._basic_max,
                           task,
                           task.ids.container.children[0].ids.name)
        task.open()

    @staticmethod
    def _basic_max(task, address, *args):
        '''Gets the values from address and returns a maximum.

        .. versionadded:: 0.1.0
        '''
        values = task.from_address(task.tablenum, address.text)
        task.set_page('Maximum', str(max(values)), 'text')

    def basic_small(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.SmallLargeLayout` that
        gets from user :ref:`Data` address and `k` variable representing the
        `k`-th value from the :ref:`Task` s output.

        .. versionadded:: 0.1.0
        '''
        widget = SmallLargeLayout()
        task = Task(title='Small', wdg=widget,
                    call=['Small', Basic.basic_small])
        task.run = partial(Basic._basic_small,
                           task,
                           task.ids.container.children[0].ids.name,
                           task.ids.container.children[0].ids.order)
        task.open()

    @staticmethod
    def _basic_small(task, address, k, *args):
        '''Gets the values from address and returns the `k`-th value from
        the ascending list of sorted values.

        .. versionadded:: 0.1.0
        '''
        values = task.from_address(task.tablenum, address.text)
        values = sorted(values)
        k = int(k.text) - 1
        try:
            task.set_page('Small ({}.)'.format(k + 1), str(values[k]), 'text')
        except IndexError:
            pass

    def basic_large(self, *args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.SmallLargeLayout` that
        gets from user :ref:`Data` address and `k` variable representing the
        `k`-th value from the :ref:`Task` s output.

        .. versionadded:: 0.1.0
        '''
        widget = SmallLargeLayout()
        task = Task(title='Large', wdg=widget,
                    call=['Large', Basic.basic_large])
        task.run = partial(Basic._basic_large,
                           task,
                           task.ids.container.children[0].ids.name,
                           task.ids.container.children[0].ids.order)
        task.open()

    @staticmethod
    def _basic_large(task, address, k, *args):
        '''Gets the values from address and returns the `k`-th value from
        the descending list of sorted values.

        .. versionadded:: 0.1.0
        '''
        values = task.from_address(task.tablenum, address.text)
        values = sorted(values, reverse=True)
        k = int(k.text) - 1
        try:
            task.set_page('Large ({}.)'.format(k + 1), str(values[k]), 'text')
        except IndexError:
            pass  # throw error k out of len(values) bounds, same for *_small

    def basic_freq(self, *args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.FreqLayout` that
        gets from user:

        * :ref:`data` address
        * type of frequency (`absolute`, `relative` or `cumulative`
        * |b|_
        * lower and upper limit (optional, by default takes `min` and `max`)

        .. versionadded:: 0.3.2
        .. versionchanged:: 0.3.8
           Switched to NumPy's `histogram` which is more flexible.
        '''
        widget = FreqLayout()
        task = Task(title='Frequency', wdg=widget,
                    call=['Frequency', Basic.basic_freq])
        task.run = partial(
            Basic._basic_freq,
            task,
            task.ids.container.children[0].ids.name,
            task.ids.container.children[0].ids.precision,
            (task.ids.container.children[0].ids.binmanager,
             task.ids.container.children[0].ids.bins,
             task.ids.container.children[0].ids.bingrid,
             task.ids.container.children[0].ids.binstr),
            (task.ids.container.children[0].ids.lowlimit,
             task.ids.container.children[0].ids.uplimit,
             task.ids.container.children[0].ids.limits_auto),
            (task.ids.container.children[0].ids.absolute,
             task.ids.container.children[0].ids.relative,
             task.ids.container.children[0].ids.cumulative))
        task.open()

    @staticmethod
    def _basic_freq_prec(input_list, precision, *args):
        if precision == '':
            return input_list
        try:
            precision = int(precision)
        except ValueError:
            raise Exception('Bad precision input!')
        if precision:
            return [round(item, precision) for item in input_list]


    @staticmethod
    def _basic_freq(task, address, precision, bins, limits, freq_type, *args):
        '''Gets the values from address and depending on the inputed bins:

        * `Count` of equal-width bins
        * `Edges` manually created edges for non-uniform bin widths. It already
          contains minimum and maximum of the values list.
        * `Calculate` uses NumPy's way for creating an optimal bin width.

        Then according to the size of bins and limits of the frequency creates
        a table for chosen types of frequency.

        If necessary, you can set the maximal amount of decimal digits for the
        frequency outputs.

        .. versionadded:: 0.3.2
        .. versionchanged:: 0.3.8
           Switch from SciPy's frequency functions to NumPy's `histogram` which
           removes `intervals` option as this is already handled by histogram.
        .. versionchanged:: 0.3.9
           Added precision.
        '''

        # input variables
        bin_type, bins, bingrid, binstr = bins
        lowlimit, uplimit, limits_auto = limits
        absolute, relative, cumulative = freq_type
        precision = precision.text

        # setting variables
        values = task.from_address(task.tablenum, address.text)
        values = [float(val) for val in values]
        values = sorted(values)
        cols = 2

        if limits_auto.active:
            lowlimit = min(values)
            uplimit = max(values)
        else:
            lowlimit = float(lowlimit.text)
            uplimit = float(uplimit.text)

        # get either int of bins or float list of edges
        bin_type = bin_type.current
        if bin_type == 'Count':
            bins = int(bins.text)
        elif bin_type == 'Edges':
            bins = [float(child.text) for child in bingrid.children]
            bins = sorted(bins)
            if bins[0] > min(values):
                bins.insert(0, min(values))
            if bins[-1] < max(values):
                bins.extend(max(values))
        elif bin_type == 'Calculate':
            bins = binstr.su

        # get base for results
        histo, edges = histogram(values, bins=bins, range=(lowlimit, uplimit))
        values_len = float(len(values))

        # set edge values
        left = []
        right = []
        for i in xrange(len(edges)):
            left.append(edges[i])
            try:
                right.append(edges[i+1])
            except IndexError:
                right.append('-')

        # "float(val)" to convert from numpy's type
        absol = []
        if absolute.active:
            absol = [float(val) for val in histo]
            absol.insert(0, 'Absolute')
            cols += 1

        relat = []
        if relative.active:
            if absol:
                relat = [val / values_len for val in absol[1:]]
            else:
                relat = [float(val) / values_len for val in histo]
            relat = Basic._basic_freq_prec(relat, precision)
            relat.insert(0, 'Relative')
            cols += 1

        cumul = []
        if cumulative.active:
            if not relative.active:
                _relat = [float(val) / values_len for val in histo]
            else:
                _relat = relat[1:]

            previous = 0
            for i in xrange(len(histo)):
                cumul.append(previous + float(_relat[i]))
                previous += float(_relat[i])
            cumul = Basic._basic_freq_prec(cumul, precision)
            cumul.insert(0, 'Cumulative')
            cols += 1

        # zipping & exporting results
        rs = []
        left.insert(0, 'Lower edge')
        right.insert(0, 'Upper edge')

        if absol and not relat and not cumul:
            rs = [r for items in zip(left, right, absol) for r in items]
        elif absol and relat and not cumul:
            rs = [r for items in zip(left, right, absol, relat) for r in items]
        elif absol and cumul and not relat:
            rs = [r for items in zip(left, right, absol, cumul) for r in items]
        elif relat and not absol and not cumul:
            rs = [r for items in zip(left, right, relat) for r in items]
        elif relat and cumul and not absol:
            rs = [r for items in zip(left, right, relat, cumul) for r in items]
        elif cumul and not absol and not relat:
            rs = [r for items in zip(left, right, cumul) for r in items]
        elif absol and relat and cumul:
            zipped = zip(left, right, absol, relat, cumul)
            rs = [r for items in zipped for r in items]

        task.set_page('Frequency', rs, 'table{}'.format(cols))

    names = (('Count', basic_count),
             ('_Count ifs', basic_countifs),
             ('Minimum', basic_min),
             ('Maximum', basic_max),
             ('Small', basic_small),
             ('Large', basic_large),
             ('Frequency', basic_freq))



import unittest

import os
import sys
import time
import sqlite3
import os.path as op
from shutil import rmtree
from functools import partial
from kivy.clock import Clock

main_path = op.dirname(op.dirname(op.abspath(__file__)))
sys.path.append(main_path)
from main import KrySA


class Test(unittest.TestCase):
    def pause(*args):
        time.sleep(0.000001)

    def run_test(self, app, *args):
        Clock.schedule_interval(self.pause, 0.000001)

        # open open_project dialog
        app.root.open_project()
        test_project = op.join(op.dirname(self.folder), 'test_Project')
        app.root.opendlg.run([op.join(test_project, 'Test.krysa')])

        self.assertTrue(app.project_exists)
        self.assertEqual(app.project_dir, test_project)
        self.assertEqual(app.project_name, 'Test')
        data = op.join(test_project, 'data')
        results = op.join(test_project, 'results')

        # set testing data
        newdata_values = [[0, u'', 0.0],
                          [0, u'', 0.0],
                          [0, u'end', 0.0],
                          [0, u'', 1.1],
                          [1, u'', 0.0]]
        newdata2_values = [[1, u'text', 1.1]]
        results_count = 2
        self.assertEqual(results_count, len(os.listdir(results)))

        # get data from krysa table
        ndv = []
        for d in app.root.tables[0][1].rv.data:
            if 'c' in d.keys():
                if d['type'] == int:
                    ndv.append(int(d['text']))
                elif d['type'] == float:
                    ndv.append(float(d['text']))
                else:
                    ndv.append(d['text'])
        ndv = [ndv[x:x + 3] for x in xrange(0, len(ndv), 3)]

        ndv2 = []
        for d in app.root.tables[1][1].rv.data:
            if 'c' in d.keys():
                if d['type'] == int:
                    ndv2.append(int(d['text']))
                elif d['type'] == float:
                    ndv2.append(float(d['text']))
                else:
                    ndv2.append(d['text'])
        ndv2 = [ndv2[x:x + 3] for x in xrange(0, len(ndv2), 3)]

        # test data
        self.assertTrue(op.exists(op.join(data, 'data.sqlite')))
        conn = sqlite3.connect(op.join(data, 'data.sqlite'))
        c = conn.cursor()
        c.execute('SELECT * FROM NewData')
        values = [item for sublist in c.fetchall() for item in sublist]
        values = [values[x:x + 3] for x in xrange(0, len(values), 3)]
        conn.close()
        print 'NewData:'
        for i, v in enumerate(values):
            # values == sql values
            self.assertEqual(v, newdata_values[i])
            # values == KrySA values
            self.assertEqual(ndv[i], newdata_values[i])
            print v

        conn = sqlite3.connect(op.join(data, 'data.sqlite'))
        c = conn.cursor()
        c.execute('SELECT * FROM NewData2')
        values = [item for sublist in c.fetchall() for item in sublist]
        values = [values[x:x + 3] for x in xrange(0, len(values), 3)]
        conn.close()
        print '\nNewData2:'
        for i, v in enumerate(values):
            self.assertEqual(v, newdata2_values[i])
            self.assertEqual(ndv2[i], newdata2_values[i])
            print v

        # test imported results
        sources = []
        imported_results = 0
        expected_files = ['000', '001']

        results = app.root.ids.results
        for child in reversed(results.children):
            if 'Page' in str(child):
                imported_results += 1
                for image in child.children[2].children[0].children:
                    sources.append(op.splitext(op.basename(image.source))[0])

        self.assertEqual(results_count, imported_results)
        self.assertEqual(sources, expected_files)

        app.stop()

    def test_file_openproject(self):
        self.path = op.dirname(op.abspath(__file__))
        if not op.exists(op.join(self.path, 'test_folder')):
            os.mkdir(op.join(self.path, 'test_folder'))
        else:
            rmtree(op.join(self.path, 'test_folder'))
            os.mkdir(op.join(self.path, 'test_folder'))
        self.folder = op.join(self.path, 'test_folder')

        app = KrySA()
        p = partial(self.run_test, app)
        Clock.schedule_once(p, .000001)
        app.run()
        rmtree(self.folder)

if __name__ == '__main__':
    unittest.main()



from . import Task, SortLayout, AppendLayout, StandLayout
from kivy.uix.tabbedpanel import TabbedPanelItem
from functools import partial


class Manipulate(object):
    '''All :ref:`Task` s categorized as being able to `manipulate` data.
    A result after manipulation is a new data.

    .. versionadded:: 0.3.5
    '''

    def manip_sort(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.SortLayout` that gets
        from user the table which will be sorted and the type of sorting
        (`Ascending` or `Descending`).

        .. versionadded:: 0.3.5
        '''
        widget = SortLayout()
        task = Task(title='Sort', wdg=widget,
                    call=['Sort', Manipulate.manip_sort])
        task.tablecls = task.app.tablecls
        task.run = partial(Manipulate._manip_sort,
                           task,
                           task.ids.container.children[0].ids.sort_type)
        task.open()

    @staticmethod
    def _manip_sort(task, sort_type, *args):
        '''Gets the values from address, sorts each column either ascending
        or descending and returns a new :mod:`main.Table`

        .. versionadded:: 0.3.5
        '''
        sort_type = 'Asc' not in sort_type.text
        from_address = task.from_address(task.tablenum, ':all', extended=True)
        values, cols, rows, labels = from_address

        # get separated cols to sort
        chunks = []
        for x in xrange(0, len(values), rows):
            chunks.append(values[x:x + rows])

        values = []
        for val in chunks:
            values.append(sorted(val, reverse=sort_type))

        # add Table
        table = task.ids.tablesel.text
        table += ' (desc)' if sort_type else ' (asc)'
        tabletab = TabbedPanelItem(text=table)
        task.app.root.ids.tabpanel.add_widget(tabletab)

        values = zip(*values)
        values = [v for vals in values for v in vals]
        task.app.root.tables.append((
            table, task.tablecls(max_cols=cols, max_rows=rows,
                                 pos=task.app.root.pos,
                                 size=task.app.root.size,
                                 values=values, labels=labels)
        ))
        tabletab.content = task.app.root.tables[-1][1]

    def manip_filter(*args):
        '''(Not yet implemented)
        '''

    def manip_append(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.AppendLayout` that gets
        from user :mod:`main.Table`, type of append and an amount of empty
        rows / cols to append.

        .. note:: Appending new columns don't work for now. When such an action
           is possible, this note will be removed.

        .. versionadded:: 0.3.6
        '''
        widget = AppendLayout()
        task = Task(title='Append', wdg=widget,
                    call=['Append', Manipulate.manip_append])
        task.tablecls = task.app.tablecls
        task.run = partial(Manipulate._manip_append,
                           task,
                           task.ids.container.children[0].ids.what,
                           task.ids.container.children[0].ids.amount,
                           task.ids.container.children[0].ids.overwrite)
        task.open()

    @staticmethod
    def _manip_append(task, append_type, amount, overwrite, *args):
        '''Gets the amount of empty rows / cols to append from user and
        either returns a new, altered :mod:`main.Table` of selected one, or
        appends directly to the selected Table.

        .. versionadded:: 0.3.6
        .. versionchanged:: 0.3.7
            Added overwriting of selected :mod:`main.Table`
        '''
        append_type = append_type.text
        overwrite = overwrite.active
        amount = int(amount.text) if amount.text else 0

        # Stop the task if no amount (or =0) is specified
        if not amount:
            return

        from_address = task.from_address(task.tablenum, ':all',
                                         extended=True)
        values, cols, rows, labels = from_address

        if append_type == 'Columns':
            cols += amount
            return
        elif append_type == 'Rows':
            # get columns
            chunks = []
            for x in xrange(0, len(values), rows):
                chunks.append(values[x:x + rows])

            # append to columns a zero value according
            # to their type int(amount)-times
            for r in range(amount):
                for chunk in chunks:
                    if isinstance(chunk[0], int):
                        chunk.append(0)
                    elif isinstance(chunk[0], float):
                        chunk.append(0.0)
                    else:
                        chunk.append(u'')

            # add Table
            tab_pos = 0
            table = task.ids.tablesel.text
            tabletab = TabbedPanelItem(text=table)
            if overwrite:
                tab_pos = task.tablenum + 1
                old_tab = task.app.root.ids.tabpanel.tab_list[tab_pos]
                task.app.root.ids.tabpanel.remove_widget(old_tab)
            else:
                table += ' (append {})'.format(str(amount))
            task.app.root.ids.tabpanel.add_widget(tabletab, tab_pos)

            # zip chunks to values, flatten values
            values = zip(*chunks)
            values = [v for vals in values for v in vals]

            # increase row count by new rows, make new table
            rows += amount
            new_table = (
                table, task.tablecls(max_cols=cols, max_rows=rows,
                                     pos=task.app.root.pos,
                                     size=task.app.root.size,
                                     values=values, labels=labels)
            )

            # place newly created table into tab's content
            if overwrite:
                task.app.root.tables[tab_pos - 1] = new_table
                tabletab.content = task.app.root.tables[tab_pos - 1][1]
            else:
                task.app.root.tables.append(new_table)
                tabletab.content = task.app.root.tables[-1][1]
        else:
            return

    def manip_split(*args):
        '''(Not yet implemented)
        '''

    def manip_merge(*args):
        '''(Not yet implemented)
        '''

    def manip_stand(*args):
        '''Opens a :mod:`tasks.Task` with a :mod:`tasks.StandLayout` that gets
        from user :mod:`main.Table`, type of append and an amount of empty
        rows / cols to append.

        .. note:: Appending new columns don't work for now. When such an action
           is possible, this note will be removed.

        .. versionadded:: 0.3.6
        '''
        widget = AppendLayout()
        task = Task(title='Append', wdg=widget,
                    call=['Append', Manipulate.manip_append])
        task.tablecls = task.app.tablecls
        task.run = partial(Manipulate._manip_append,
                           task,
                           task.ids.container.children[0].ids.what,
                           task.ids.container.children[0].ids.amount,
                           task.ids.container.children[0].ids.overwrite)
        task.open()

    @staticmethod
    def _manip_stand(task, append_type, amount, overwrite, *args):
        '''Gets the amount of empty rows / cols to append from user and
        either returns a new, altered :mod:`main.Table` of selected one, or
        appends directly to the selected Table.

        .. versionadded:: 0.3.6
        .. versionchanged:: 0.3.7
            Added overwriting of selected :mod:`main.Table`
        '''

    names = (('Sort', manip_sort),
             ('_Filter', manip_filter),
             ('Append', manip_append),
             ('_Split', manip_split),   # split into two data(columns, rows)
             ('_Merge', manip_merge),
             ('_Standardize', manip_stand))



from . import Task, AvgsLayout, AddressLayout
from functools import partial, reduce
from collections import Counter
import operator
import math


class Avgs(object):
    '''All :ref:`Task` s categorized as `averages` under one roof.

    .. versionadded:: 0.2.4
    '''

    def avgs_gen(*args):
        '''Generalized mean:

        .. math::
           \\big(\\frac{1}{n}\\sum_{i=1}^n  x_{i}^{p} \\big)^\\frac{1}{p},
           where:

        * p == -1: harmonic
        * p == 0: geometric
        * p == 1: arithmetic
        * p == 2: quadratic
        * p == 3: cubic
        * etc...

        .. versionadded:: 0.2.4
        '''
        widget = AvgsLayout()
        task = Task(title='Count', wdg=widget,
                    call=['Count', Avgs.avgs_gen])
        task.run = partial(Avgs._avgs_gen,
                           task,
                           task.ids.container.children[0].ids.name,
                           task.ids.container.children[0].ids.power)
        task.open()

    @staticmethod
    def _avgs_gen(task, address, p, *args):
        '''Gets the values from address and depending on `p` (power) value
        returns either exceptional case for `p == 0` (geometric mean), or
        value from the generalized mean's formula.

        .. versionadded:: 0.2.4
        '''
        if p == '-0':
            p = 0.0
        else:
            p = float(p.text)

        values = task.from_address(task.tablenum, address.text)
        if p in [0, -0]:
            multiples = reduce(operator.mul, values, 1)
            result = math.pow(multiples, 1 / float(len(values)))
            p = 'geometric'
        else:
            values = [math.pow(val, p) for val in values]
            upper = math.fsum(values) / float(len(values))
            result = math.pow(upper, 1 / float(p))
        task.set_page('Generalized mean({})'.format(p), str(result), 'text')

    def avgs_inter(*args):
        '''(Not yet implemented)
        '''

    def avgs_mid(*args):
        '''(Not yet implemented)
        '''

    def avgs_trim(*args):
        '''(Not yet implemented)
        '''

    def avgs_median(*args):
        '''Median:

        .. math::
            \\tilde x = \\left \\{ \\begin {array}{lr}
            \\frac {n}{2} \in \\mathbb{N}:&
            \\frac {x_{\\frac {n}{2}} + x_{\\frac {n}{2}+1}}{2}
            \\\\
            \\frac {n+1}{2} \in \\mathbb{N}:& x_{\\lceil \\frac {n}{2} \\rceil}
            \\end{array} \\right.

        .. versionadded:: 0.3.10
        '''
        widget = AddressLayout()
        task = Task(title='Median', wdg=widget,
                    call=['Median', Avgs.avgs_median])
        task.run = partial(Avgs._avgs_median,
                           task,
                           task.ids.container.children[0].ids.name)
        task.open()

    @staticmethod
    def _avgs_median(task, address, *args):
        values = task.from_address(task.tablenum, address.text)
        print values
        length = len(values)
        pos = int(round(length / 2.0))
        if length == 1:
            result = values[0]
        elif length % 2:
            result = values[pos - 1]
        else:
            result = int(values[pos - 1] + values[pos]) / 2.0
        task.set_page('Median', str(result), 'text')

    def avgs_mode(*args):
        '''Mode returns the most common value from the list of values.
        If there's more than a single value with the same amount of
        occurency, all values with the same occurency are returned.

        .. versionadded:: 0.3.10
        '''
        widget = AddressLayout()
        task = Task(title='Mode', wdg=widget,
                    call=['Mode', Avgs.avgs_mode])
        task.run = partial(Avgs._avgs_mode,
                           task,
                           task.ids.container.children[0].ids.name)
        task.open()

    @staticmethod
    def _avgs_mode(task, address, *args):
        values = task.from_address(task.tablenum, address.text)
        length = len(values)
        data = Counter(values)
        occurs = data.most_common(length)
        results = []
        for oc in occurs:
            if not results:
                results.append(oc)
            else:
                if results[-1][1] == oc[1]:
                    results.append(oc)
        results = [str(r[0]) for r in results]
        task.set_page('Mode', ', '.join(results), 'text')

    names = (('Generalized', avgs_gen),
             ('_Interquartile', avgs_inter),
             ('_Midrange', avgs_mid),
             ('_Trimmed', avgs_trim),
             ('Median', avgs_median),
             ('Mode', avgs_mode))



'''
Tests run independently on each other, but not in a single python interpreter.
After each test a fresh `python` is required (it won't run as casual suite),
so either run them like this one by one::

    python test_<something>.py

or use `KivyUnitTest <https://github.com/KeyWeeUsr/KivyUnitTest>`_ to do it
instead of you and better.
'''



#!/usr/bin/env python

import os
import sys
import os.path as op
from distutils import dir_util
from distutils import sysconfig
from setuptools import setup, find_packages

name = 'krysa'
files = [name, ]

root = op.dirname(op.abspath(__file__)) + op.sep

# remove all *.pyc files here before packaging
if 'sdist' in sys.argv:
    rem_files = []
    for path, folders, filenames in os.walk(root):
        for file in filenames:
            rem_files.append(op.join(path, file))
    for file in rem_files:
        if '.pyc' in file:
            print('Removing {}'.format(file))
            os.remove(file)

# get version from main
with open(op.join(root, 'krysa', 'main.py')) as f:
    for i, line in enumerate(f):
        if i == 2:
            version = line[len('# Version: '):-1]
        elif i > 2:
            break

# check version format
ver_split = version.split('.')
num_check = [int(n) for n in ver_split]
if len(ver_split) != 3:
    raise Exception('No correct version! ( {0} )'.format(version))

setup(
    name=name,
    packages=find_packages(),
    include_package_data=True,
    install_requires=['kivy', 'numpy', 'scipy', 'matplotlib'],
    version=version,
    description='KrySA - Statistical analysis for rats',
    author='Peter Badida',
    author_email='keyweeusr@gmail.com',
    url='https://github.com/KeyWeeUsr/KrySA',
    download_url='https://github.com/KeyWeeUsr/KrySA/tarball/' + version,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: Android',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    keywords=['krysa', 'statistics', 'kivy'],
    license="GNU General Public License v3 (GPLv3)",
)



import unittest

import os
import sys
import time
import sqlite3
import os.path as op
from shutil import rmtree
from functools import partial
from kivy.clock import Clock

main_path = op.dirname(op.dirname(op.abspath(__file__)))
sys.path.append(main_path)
from main import KrySA, ErrorPop


class Test(unittest.TestCase):
    def pause(*args):
        time.sleep(0.000001)

    def run_test(self, app, *args):
        Clock.schedule_interval(self.pause, 0.000001)

        # open New -> Project popup, set inputs
        app.root._new_project()
        app.root.savedlg.view.selection = [self.folder, ]
        app.root.savedlg.ids.name.text = 'Test.krysa'
        app.root.savedlg.run([self.folder, ], 'Test.krysa')
        project_folder = op.join(self.path, 'test_folder', 'Test')
        data = op.join(project_folder, 'data')
        results = op.join(project_folder, 'results')

        # open New -> Data popup, set inputs
        app.root._new_data()
        new_data = app.root.wiz_newdata.ids.container.children[0]
        new_data.ids.table_name.text = 'NewData'

        # set columns for new data
        cols = new_data.ids.columns.children
        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'INTEGER'
        vals = cols[0].ids.vals.children
        for i in range(5):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 4:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = '1'

        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'TEXT'
        vals = cols[0].ids.vals.children
        for i in range(3):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 2:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = 'end'

        new_data.ids.columnadd.dispatch('on_release')
        cols[0].ids.colname.text += str(len(cols))
        cols[0].ids.coltype.text = 'REAL'
        vals = cols[0].ids.vals.children
        for i in range(4):
            cols[0].ids.valadd.dispatch('on_release')
            if i != 3:
                vals[0].ids.value.text = ''
            else:
                vals[0].ids.value.text = '1.1'
        new_data = app.root.wiz_newdata.run()

        # test project variables
        self.assertTrue(app.project_exists)
        self.assertFalse(app.project_dir is '')
        self.assertFalse(app.project_name is '')

        # test closed project
        app.root.close_project()

        self.assertFalse(app.project_exists)
        self.assertTrue(app.project_dir is '')
        self.assertTrue(app.project_name is '')
        self.assertFalse(app.root.tables)
        self.assertFalse(app.root.ids.flow.children)
        self.assertTrue(len(app.root.ids.tabpanel.tab_list) is 1)

        app.stop()

    def test_file_new(self):
        self.path = op.dirname(op.abspath(__file__))
        if not op.exists(op.join(self.path, 'test_folder')):
            os.mkdir(op.join(self.path, 'test_folder'))
        else:
            rmtree(op.join(self.path, 'test_folder'))
            os.mkdir(op.join(self.path, 'test_folder'))
        self.folder = op.join(self.path, 'test_folder')

        app = KrySA()
        p = partial(self.run_test, app)
        Clock.schedule_once(p, .000001)
        app.run()
        rmtree(self.folder)

if __name__ == '__main__':
    unittest.main()
